      module mstack
     
         ! vars
         
         integer, parameter :: stack_inc = 1000
         
         type rstack
            double precision, allocatable :: rdata(:,:)
            integer :: ptr
            integer :: sze
         end type
         
         type istack
            integer, allocatable :: idata(:,:)
            integer :: ptr
            integer :: sze
         end type
         
         !interfaces
     
         interface get_el !get element
            module procedure get_el_i, get_el_r
         end interface
         
         interface get_len !get length
            module procedure get_len_r, get_len_i
         end interface
         
         interface init !initialise
            module procedure init_r, init_i
         end interface
         
         interface destroy !destroy...
            module procedure destroy_r, destroy_i
         end interface
         
         interface rep_el !replace element
            module procedure rep_el_r, rep_el_i
         end interface
         
         interface pop !delete last element
            module procedure pop_r, pop_i
         end interface
         
         interface push !insert element at the end
            module procedure push_r, push_i
         end interface
         
         interface empty !return if element is empty
            module procedure empty_r, empty_i
         end interface
         
         interface wout !write out stack
            module procedure wout_r, wout_i
         end interface
         
         interface clear !clear stack
            module procedure clear_r, clear_i
         end interface
         
         interface getmax !get index of largest element
            module procedure getmax_r, getmax_i
         end interface
         
         interface push_single !insert element if not already present
            module procedure push_single_r, push_single_i
         end interface
     
         !functions
         contains!new stack:
         function get_len_r(this)
            integer :: get_len_r
            type(rstack) :: this
            get_len_r = this%ptr
            return
         end function

         function get_len_i(this)
            integer :: get_len_i
            type(istack) :: this
            get_len_i = this%ptr
            return
         end function

         subroutine init_r(this, sze0)
            integer, intent(in) :: sze0
            type(rstack) :: this
            allocate(this%rdata(stack_inc,sze0))
            this%ptr = 0
            this%sze = sze0
         end subroutine

         subroutine init_i(this, sze0)
            integer, intent(in) :: sze0
            type(istack) :: this
            allocate(this%idata(stack_inc,sze0))
            this%ptr = 0
            this%sze = sze0
         end subroutine
         
         subroutine destroy_r(this)
            type(rstack) :: this
            if(allocated(this%rdata))then
               deallocate(this%rdata)
            endif
         end subroutine
         
         subroutine destroy_i(this)
            type(istack) :: this
            if(allocated(this%idata))then
               deallocate(this%idata)
            endif
         end subroutine

         subroutine rep_el_r(this,i,rin)
            type(rstack) :: this
            integer, intent(in) :: i
            double precision, intent(in) :: rin(this%sze)
            if(i.le.this%ptr)then
               this%rdata(i,:) = rin
            end if
         end subroutine

         subroutine rep_el_i(this,i,iin) 
            type(istack) :: this
            integer, intent(in) :: i
            integer, intent(in) :: iin(this%sze)
            if(i.le.this%ptr)then
               this%idata(i,:) = iin
            end if
         end subroutine
         
         function get_el_r(this, i)
            integer, intent(in) :: i
            type(rstack) :: this
            double precision :: get_el_r(size(this%rdata,2))
            if(i.le.this%ptr)then
               get_el_r(:) = this%rdata(i,:)
            end if
            return
         end function

         function get_el_i(this, i)
            integer, intent(in) :: i
            type(istack) :: this
            integer :: get_el_i(size(this%idata,2))
            if(i.le.this%ptr)then
               get_el_i(:) = this%idata(i,:)
            end if
            return
         end function
         
         subroutine pop_r(this)
            type(rstack) :: this
            if(this%ptr.ge.1)then
               this%ptr = this%ptr - 1
            end if
         end subroutine
         
         subroutine pop_i(this)
            type(istack) :: this
            if(this%ptr.ge.1)then
               this%ptr = this%ptr - 1
            end if
         end subroutine
         
         subroutine push_r(this, rin) 
            type(rstack) :: this
            double precision, intent(in) :: rin(this%sze)
            double precision, allocatable :: rtmp(:,:)
            if(this%ptr.ge.size(this%rdata,1))then !resize
               allocate(rtmp(size(this%rdata,1),this%sze))
               rtmp(:,:) = this%rdata(:,:)
               deallocate(this%rdata)
               allocate(this%rdata(size(rtmp,1)+stack_inc,size(rtmp,2)))
               this%rdata(1:size(rtmp,1),:) = rtmp(:,:)
               deallocate(rtmp)
            endif
            this%ptr = this%ptr + 1
            this%rdata(this%ptr,:) = rin(:)
         end subroutine
         
         subroutine push_i(this, iin)
            type(istack) :: this
            integer, intent(in) :: iin(this%sze)
            integer, allocatable :: itmp(:,:)
            if(this%ptr.ge.size(this%idata,1))then !resize
               allocate(itmp(size(this%idata,1),this%sze))
               itmp(:,:) = this%idata(:,:)
               deallocate(this%idata)
               allocate(this%idata(size(itmp,1)+stack_inc,size(itmp,2)))
               this%idata(1:size(itmp,1),:) = itmp(:,:)
               deallocate(itmp)
            endif
            this%ptr = this%ptr + 1
            this%idata(this%ptr,:) = iin(:)
         end subroutine
         
         function empty_r(this)
            type(rstack) :: this
            logical empty_r
            empty_r = .FALSE.
            if(this%ptr.le.0)then
               empty_r = .TRUE.
            end if
            return
         end function
         
         function empty_i(this)
            type(istack) :: this
            logical empty_i
            empty_i = .FALSE.
            if(this%ptr.le.0)then
               empty_i = .TRUE.
            end if
            return
         end function
         
         subroutine wout_r(this)
            type(rstack) :: this
            write(*,*) this%rdata(1:this%ptr,:)
         end subroutine
         
         subroutine wout_i(this)
            type(istack) :: this
            write(*,*) this%idata(1:this%ptr,:)
         end subroutine
         
         subroutine clear_r(this)
            type(rstack) :: this
            this%ptr = 0
         end subroutine

         subroutine clear_i(this)
            type(istack) :: this
            this%ptr = 0
         end subroutine
         
         function getmax_i(this)
            type(istack) :: this
            integer :: getmax_i, ret(1)
            ret = maxloc(this%idata(1:this%ptr,1))
            getmax_i = ret(1)
            return
         end function
         
         function getmax_r(this)
            type(rstack) :: this
            integer :: getmax_r, ret(1)
            ret = maxloc(this%rdata(1:this%ptr,1))
            getmax_r = ret(1)
            return
         end function
         
         subroutine push_single_r(this, rin)
            type(rstack) :: this
            double precision, intent(in) :: rin(this%sze)
            do i = 1, this%ptr
               if(this%rdata(i,1).eq.rin(1)) return
            end do
            call push_r(this,rin)
         end subroutine
         
         subroutine push_single_i(this, iin)
            type(istack) :: this
            integer, intent(in) :: iin(this%sze)
            do i = 1, this%ptr
               if(this%idata(i,1).eq.iin(1)) return
            end do
            call push_i(this,iin)
         end subroutine
      end module
